import argparse
import os
import sys
from git import Repo
from git.exc import InvalidGitRepositoryError, NoSuchPathError
from git_workflow.__about__ import __version__
from git_workflow.utils import cmd, repository
from git_workflow.workflow import add_command_subparsers, run_command


def get_generic_parent_parser():
    """Returns a generic parent ArgumentParser with --help, --version, and
    --verbose args

    Note: parsers that have this as a parent should be initialized with
    add_help=False. This parser overrides the default help arg so it can be
    nested under the 'General' group in help message output

    :return: ArgumentParser to use as parent
    """
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group('General')
    # Overriding the default -h arg so it can be nested under the 'General'
    # heading in help message. Parsers that have this as a parent should be
    # initialized with add_help=False
    group.add_argument('-h', '--help', action='help',
                       help='Show this help message and exit')
    group.add_argument('-V', '--version', action='version',
                       version='%(prog)s ' + __version__,
                       help='Show version number and exit')
    # TODO: Implement verbose
    # group.add_argument('-v', '--verbose', type=int, choices=range(0,3),
    #                    nargs='?', default=1, const=2,
    #                    help='Set verbosity level')
    return parser

def get_parser(prog=None):
    """Returns ArgumentParser for main"""
    generic_parent_parser = get_generic_parent_parser()
    parser = argparse.ArgumentParser(parents=[generic_parent_parser],
                                     add_help=False,
                                     prog=prog)
    subparsers = parser.add_subparsers(title='Commands',
                                       description=f"Run '{parser.prog} <command> --help' for details",
                                       dest='command', metavar='<command>')
    add_command_subparsers(subparsers, generic_parent_parser)
    return parser


def main():
    # Check installed git version
    # TODO allow limited functionality for lower git versions in future update?
    try:
        min_git_version_met = repository.verify_git_version()
    except Exception as e:
        cmd.print_error(e)
        return
    # Initialize Repo object
    repo = None
    try:
        repo = Repo(os.getcwd(), search_parent_directories=True)
    except InvalidGitRepositoryError as e:
        cmd.print_error('No git repo found: {}'.format(e))
    except NoSuchPathError as e:
        cmd.print_error('Invalid path: {}'.format(e))
    finally:
        if repo is None:
            return
    # Argument Parser
    parser = get_parser()
    try:
        run_command(repo, parser)
    except KeyboardInterrupt:
        print('')
        sys.exit(0)
    except Exception as e:
        cmd.print_error(e)
        sys.exit(1)


if __name__ == '__main__':
    main()



"""The workflow scripts"""
from .start_branch import StartBranch
from .set_template import SetTemplate
from .unset_template import UnsetTemplate
from .finish_branch import FinishBranch
from .cleanup import Cleanup

#: Maps command names to WorkflowBase subclasses
commands = {
    StartBranch.command: StartBranch,
    FinishBranch.command: FinishBranch,
    SetTemplate.command: SetTemplate,
    UnsetTemplate.command: UnsetTemplate,
    Cleanup.command: Cleanup,
}


def add_command_subparsers(subparsers, generic_parent_parser):
    """Add subparsers for each workflow command.

    :param subparsers: Subparsers object
    :param generic_parent_parser: Instance of argparse.ArgumentParser used
        as a parent for top level parser
    """
    for commandClass in commands.values():
        commandClass.add_subparser(subparsers, generic_parent_parser)


def run_command(repo, parser):
    """Run a workflow command.

    :param repo: Repo object
    :param parser: ArgumentParser object
    """
    parsed_args = parser.parse_args()
    command_class = commands.get(parsed_args.command, False)
    if not command_class:
        # exit_code = 1 # TODO implement? or raise specific exception
        parser.print_help()
    else:
        command = command_class(repo, parser, parsed_args=parsed_args)
        command.run()



import glob
import os
import re
from git_workflow.utils import cmd
from .base import WorkflowBase
from .unset_template import UnsetTemplate


class Cleanup(WorkflowBase):
    """\
    Tidy up workflow-related files and configs.

    This command will first list all branches and files that will be affected.
    By default, it will then prompt for confirmation unless ``--force`` is
    specified. Once confirmed, this command will:

    - Unset the commit templates of all configured branches
    - Remove any orphan commit template files that have no associated branch

    By default, the current branch is omitted from cleanup unless
    ``--include-current-branch`` is specified.
    """

    command = 'cleanup'
    description = 'Tidy up workflow-related files and configs.'
    configs_used = ['cleanupConfirmationPrompt']

    @classmethod
    def add_subparser(cls, subparsers, generic_parent_parser):
        cleanup_subparser = subparsers.add_parser(
            cls.command, description=cls.description, help=cls.description,
            parents=[generic_parent_parser], add_help=False
        )
        # Cleanup Options
        cleanup_args = cleanup_subparser.add_argument_group(
            'Cleanup Options'
        )
        cleanup_args.add_argument(
            '-B', '--include-current-branch', help='Unset template for current branch too',
            action='store_true', default=False
        )
        cleanup_args.add_argument(
            '-o', '--orphans-only', help='Only clean up templates without a branch',
            action='store_true', default=False
        )
        # Confirmation prompt
        confirmation_args = cleanup_subparser.add_argument_group(
            'Confirmation Prompt Arguments',
            'Override workflow.cleanupConfirmationPrompt config.'
        )
        confirmation_group = confirmation_args.add_mutually_exclusive_group()
        confirmation_group.add_argument(
            '-f', '--force', help='Skip confirmation prompt (if configured)',
            dest='confirm', action='store_false', default=None
        )
        confirmation_group.add_argument(
            '-c', '--confirmation', help='Prompt for confirmation before cleaning up templates',
            dest='confirm', action='store_true', default=None
        )

    def get_args(self):
        args = {}
        args['include_current_branch'] = self.parsed_args.include_current_branch
        args['orphans_only'] = self.parsed_args.orphans_only
        # Default to config value unless otherwise specified
        args['confirm'] = (self.configs.CLEANUP_CONFIRMATION_PROMPT
                           if self.parsed_args.confirm is None else
                           self.parsed_args.confirm)
        return args

    def run(self):
        args = self.get_args()
        targets = self.find_cleanup_targets()
        orphans = targets.pop('orphans', [])
        # Pop current branch's configs if we shouldn't include it
        if not args['include_current_branch']:
            current_branch = targets.pop(self.repo.active_branch.name, None)
        # Otherwise keep track of it for messaging
        else:
            current_branch = targets.get(self.repo.active_branch.name, None)
        # Output: Configured Templates (unless --orphans-only)
        if not args['orphans_only']:
            # Case 1: We have branches to clean
            if targets:
                self.print('Commit templates will be unset for the following branches:',
                           '',
                           *targets.keys(),
                           '')
                if args['include_current_branch'] and current_branch:
                    self.print_warning('This includes the branch you are currently on!',
                                       '')
            # Case 2: Nothing to clean but current branch has a template
            elif current_branch:
                self.print('Only branch with a template configured is the current branch.')
                self.print('To include the current branch, use --include-current-branch argument.',
                           '')
        # Output: Orphans
        if orphans:
            self.print('The following commit templates do not have an associated branch and will be deleted:',
                       '',
                       *orphans,
                       '')
        # Output: Nothing to clean up
        if (not targets or args['orphans_only']) and (not orphans):
            self.print('Nothing to clean up.')
            return
        # Confirmation
        if args['confirm']:
            confirmation = cmd.prompt(
                'Confirm (y/n)',
                'Would you like to continue?',
                default_val='n', validate_function=cmd.validate_yn
            )
            if not confirmation:
                return
        # Unset Configured Templates
        if not args['orphans_only'] and targets:
            self.print('Unsetting configured templates...')
            for branch_name in targets.keys():
                unset_template_parsed_args = self.parser.parse_args([UnsetTemplate.command, branch_name, '--force'])
                unset_template = UnsetTemplate(self.repo, self.parser,
                                               parsed_args=unset_template_parsed_args,
                                               # TODO match verbosity if > 1?
                                               verbosity=0)
                unset_template.run()
                self.print_success(f'{branch_name} template unset.')
            self.print('')
        # Delete Orphans
        if orphans:
            self.print('Removing orphan templates...')
            repo_root_dir = os.path.dirname(self.repo.git_dir)
            for orphan in orphans:
                orphan_path = os.path.join(repo_root_dir, orphan)
                # NOTE: Probably don't need this exists() check since find_cleanup_targets()
                # builds this list based on the files it finds, but doing it just to be safe
                if os.path.exists(orphan_path):
                    os.remove(orphan_path)
                    if not os.path.exists(orphan_path):
                        self.print_success(f'Deleted {orphan}.')
                    else:
                        self.print_warning(f'Unable to delete {orphan}.')
            self.print('')

    # Helper Methods

    def find_cleanup_targets(self):
        """Get configured and orphaned commit templates.

        :return: Dictionary with keys for each branch that map to a dictionary
            with keys 'config' and 'template' (the respective branch's config
            file and configured template). If there are orphaned commit
            templates, there will also be a key 'orphans' that's mapped to a
            list of orphaned commit templates.
        """
        # Get all branch-specific config paths
        config_expr = 'includeif\.onbranch:(.*)\.path'
        config_matches = self.configs.call_config_command(
            config_expr, file=self.configs.CONFIG_PATH, get_regexp=True
        )
        if config_matches is not None:
            config_matches = config_matches.split('\n')
        else:
            config_matches = []
        # Map branch names to their config files and configured templates
        expr = r'includeif\.onbranch:(.*)\.path (.*)'
        targets = {}
        # Used to keep track of all configured commit templates so orphaned ones can be determined
        configured_commit_templates = []
        for line in config_matches:
            match = re.match(expr, line)
            if match:
                branch_name = match.group(1)
                branch_config_file = match.group(2)
                branch_commit_template = self.configs.get_config(
                    'commit.template', file=os.path.join(self.repo.git_dir, branch_config_file)
                )
                targets[branch_name] = {
                    'config': branch_config_file,
                    'template': branch_commit_template,
                }
                configured_commit_templates.append(branch_commit_template)
        # Find orphaned templates
        repo_root_dir = os.path.dirname(self.repo.git_dir)
        all_commit_templates = [
            os.path.basename(path)
            for path in glob.glob(os.path.join(repo_root_dir, '.gitmessage_local*'))
        ]
        orphan_commit_templates = [
            template for template in all_commit_templates
            if template not in configured_commit_templates
        ]
        if orphan_commit_templates:
            targets['orphans'] = orphan_commit_templates
        return targets



"""Package info"""

__all__ = [
    '__command__', '__version__', '__min_python_version__', '__min_git_version__',
    '__url__',
]

#: Base command
__command__ = 'workflow'
#: Package version
__version__ = '0.1.2'
#: Minimum required python version
__min_python_version__ = '3.6'
#: Minimum required git version for advanced features
__min_git_version__ = 2.23
#: Project URL
__url__ = 'https://github.com/connordelacruz/git-workflow'



"""Utilities for interacting with git configs."""
import textwrap
from git.exc import GitCommandError
from . import repository


class Configs:
    """Object with variables representing git configs"""

    # Possible values for the --type argument in git config command.
    TYPE_BOOL = 'bool'
    TYPE_INT = 'int'
    TYPE_PATH = 'path'
    TYPE_EXPIRY_DATE = 'expiry-date'
    TYPE_COLOR = 'color'
    # Custom data types (used in convert_config_value())
    # Space separated list:
    DATA_TYPE_LIST = 'list'

    def __init__(self, repo, debug=False, no_init=False):
        self.repo = repo
        self.debug = debug
        # Make sure repo is initialized before proceeding
        if not no_init:
            repository.initialize(self.repo)

        # CONFIGS ==============================================================
        # Internal -------------------------------------------------------------
        # configPath
        self.CONFIG_PATH_DOC = (
            'Path to workflow config for this repo.'
        )
        self.CONFIG_PATH = self.get_workflow_config('configpath')

        # User Details ---------------------------------------------------------
        # initials
        self.INITIALS_DOC = textwrap.dedent(
            '''\
            The user's initials.

            If set, ``workflow start`` will skip the prompt for your initials and use this value.

            **E.g.:** To set your initials to "cd":

            ::

                git config --global workflow.initials cd
            ''')
        self.INITIALS = self.get_workflow_config('initials')

        # Branches -------------------------------------------------------------
        # baseBranch
        self.BASE_BRANCH_DOC = textwrap.dedent(
            '''\
            **Default:** ``master``

            Branch to use as a base when creating a new branch using ``workflow
            start``.

            **E.g.:** To base branches off of ``develop``:

            ::

                git config workflow.baseBranch develop
            ''')
        self.BASE_BRANCH = self.get_workflow_config(
            'baseBranch', default='master'
        )
        # badBranchNamePatterns
        self.BAD_BRANCH_NAME_PATTERNS_DOC = textwrap.dedent(
            '''\
            Set to a **space-separated** string of phrases or patterns that
            should not appear in a standard branch name. If set, ``workflow
            start`` will check for these before attempting to create a new
            branch.

            **E.g.:** if standard branch names shouldn't include the words
            ``-web`` or ``-plugins``:

            ::

                git config workflow.badBranchNamePatterns "-web -plugins"
            ''')
        self.BAD_BRANCH_NAME_PATTERNS = self.get_workflow_config(
            'badBranchNamePatterns',
            data_type=self.DATA_TYPE_LIST
        )

        # Commit Templates -----------------------------------------------------
        # TODO Document examples?
        # commitTemplateFormat
        self.COMMIT_TEMPLATE_FORMAT_DOC = textwrap.dedent(
            '''\
            **Default:** ``'[{ticket}] '``

            Format of commit template body. Supports the following placeholders:

              - ``{ticket}``: Replaced with ticket number
              - ``{branch}``: Replaced with branch name
              - ``{initials}``: Replaced with user initials (if configured)
            ''')
        self.COMMIT_TEMPLATE_FORMAT = self.get_workflow_config(
            'commitTemplateFormat', default='[{ticket}] '
        )
        # commitTemplateFilenameFormat
        self.COMMIT_TEMPLATE_FILENAME_FORMAT_DOC = textwrap.dedent(
            '''\
            **Default:** ``'{ticket}_{branch}'``

            Format of commit template filenames. Supports same placeholders as
            ``workflow.commitTemplateFormat``.

            **NOTE:** Resulting filenames will always begin with
            ``'.gitmessage_local_'``.
            ''')
        self.COMMIT_TEMPLATE_FILENAME_FORMAT = self.get_workflow_config(
            'commitTemplateFilenameFormat', default='{ticket}_{branch}'
        )

        # Ticket Numbers -------------------------------------------------------
        # TODO Document examples?
        # ticketInputFormatRegex
        self.TICKET_INPUT_FORMAT_REGEX_DOC = textwrap.dedent(
            '''\
            **Default:** ``'[a-zA-Z]+-[0-9]+'``

            Regex representing the format of a valid ticket number. Default
            format is 1 or more letters, then a hyphen, then 1 or more numbers.
            To allow any format, set to ``'.*'``.
            ''')
        self.TICKET_INPUT_FORMAT_REGEX = self.get_workflow_config(
            'ticketInputFormatRegex', default='[a-zA-Z]+-[0-9]+'
        )
        # ticketFormatCapitalize
        self.TICKET_FORMAT_CAPITALIZE_DOC = textwrap.dedent(
            '''\
            **Default:** ``true``

            If ``true``, letters in the ticket number will be capitalized after
            validation.
            ''')
        self.TICKET_FORMAT_CAPITALIZE = self.get_workflow_config(
            'ticketFormatCapitalize', default=True,
            config_type=self.TYPE_BOOL
        )

        # Confirmation Prompts -------------------------------------------------
        # finishBranchConfirmationPrompt
        self.FINISH_BRANCH_CONFIRMATION_PROMPT_DOC = textwrap.dedent(
            '''\
            **Default:** ``true``

            If ``true``, ``workflow finish`` will prompt for confirmation
            before unsetting unless ``-f`` is specified. If ``false``, will
            not prompt for confirmation unless ``-c`` is specified.
            ''')
        self.FINISH_BRANCH_CONFIRMATION_PROMPT = self.get_workflow_config(
            'finishBranchConfirmationPrompt', default=True,
            config_type=self.TYPE_BOOL
        )
        # unsetTemplateConfirmationPrompt
        self.UNSET_TEMPLATE_CONFIRMATION_PROMPT_DOC = textwrap.dedent(
            '''\
            **Default:** ``true``

            If ``true``, ``workflow unset-template`` will prompt for
            confirmation before unsetting unless ``-f`` is specified. If
            ``false``, will not prompt for confirmation unless ``-c`` is
            specified.
            ''')
        self.UNSET_TEMPLATE_CONFIRMATION_PROMPT = self.get_workflow_config(
            'unsetTemplateConfirmationPrompt', default=True,
            config_type=self.TYPE_BOOL
        )
        # cleanupConfirmationPrompt
        self.CLEANUP_CONFIRMATION_PROMPT_DOC = textwrap.dedent(
            '''\
            **Default:** ``true``

            If ``true``, ``workflow cleanup`` will prompt for confirmation 
            before cleaning unless ``-f`` is specified. If ``false``, will not 
            prompt for confirmation unless ``-c`` is specified.
            ''')
        self.CLEANUP_CONFIRMATION_PROMPT = self.get_workflow_config(
            'cleanupConfirmationPrompt', default=True,
            config_type=self.TYPE_BOOL
        )

        # END CONFIGS ==========================================================

    def get_workflow_config(self, key, default=None,
                            config_type=None, data_type=None,
                            section='workflow'):
        """Retrieve a git config value.

        :param key: The config to retrieve
        :param default: (Optional) Value to return if not configured

        :param config_type: (Optional) Use for --type argument when calling git
            config command
        :param data_type: (Default: config_type or None) Passed to
            convert_config_value(), will alter the type of the returned value

        :param section: (default: 'workflow') Section config key is defined in

        :return: The config value if it exists, otherwise the value of default,
            or None if no default is specified. If data_type is set, will be
            passed through convert_config_value()
        """
        value = default
        config = f'{section}.{key}' if section else key
        if data_type is None and config_type is not None:
            data_type = config_type
        default_arg = None if default is None else str(default)
        value = self.get_config(config,
                                default=default_arg, type=config_type)
        if value is not None and data_type:
            value = self.convert_config_value(value, data_type)
        # Show config name and value in debug mode
        if self.debug:
            print(f'{config}: "{value}"')
        return value

    def call_config_command(self, *args, **kwargs):
        """Wrapper around repo.git.config(). Catches GitCommandError and
        returns None if unsuccessful. Passes all positional and keyword
        arguments to config().
        """
        value = None
        try:
            value = self.repo.git.config(*args, **kwargs)
        except GitCommandError:
            pass
        return value

    def get_config(self, *args, **kwargs):
        return self.call_config_command(*args, get=True, **kwargs)

    def unset_config(self, *args, **kwargs):
        return self.call_config_command(*args, unset=True, **kwargs)

    @classmethod
    def convert_config_value(cls, string_val, data_type):
        """Convert a string value returned from git config command to a python
        data type.

        :param string_val: The result of the call to git config command (should be
            string)
        :param data_type: TYPE_ or DATA_TYPE_ constant defining what type of data
            this should be converted to

        :return: The converted value
        """
        converted_val = string_val
        if data_type == cls.TYPE_BOOL:
            converted_val = string_val == 'true'
        elif data_type == cls.TYPE_INT:
            converted_val = int(string_val)
        elif data_type == cls.DATA_TYPE_LIST:
            converted_val = string_val.split()
        return converted_val



import os
from git import GitCommandError
from git_workflow.utils import cmd
from .base import WorkflowBase


class UnsetTemplate(WorkflowBase):
    """\
    Remove commmit template for a branch.

    By default, this command will prompt for confirmation before removing the
    commit template unless ``--force`` is specified.
    """

    command = 'unset-template'
    description = 'Remove commit template for a branch.'
    configs_used = ['unsetTemplateConfirmationPrompt']

    @classmethod
    def add_subparser(cls, subparsers, generic_parent_parser):
        unset_commit_template_subparser = subparsers.add_parser(
            cls.command, description=cls.description, help=cls.description,
            parents=[generic_parent_parser], add_help=False
        )
        # Specify branch
        positional_args = unset_commit_template_subparser.add_argument_group(
            'Positional Arguments'
        )
        positional_args.add_argument(
            'branch', metavar='<branch>', nargs='?',
            help='Branch to unset template for (default: current)',
            default=None
        )
        # Confirmation prompt
        confirmation_args = unset_commit_template_subparser.add_argument_group(
            'Confirmation Prompt Arguments',
            'Override workflow.unsetTemplateConfirmationPrompt config.'
        )
        confirmation_group = confirmation_args.add_mutually_exclusive_group()
        confirmation_group.add_argument(
            '-f', '--force', help='Skip confirmation prompt (if configured)',
            dest='confirm', action='store_false', default=None
        )
        confirmation_group.add_argument(
            '-c', '--confirmation', help='Prompt for confirmation before unsetting',
            dest='confirm', action='store_true', default=None
        )

    def get_args(self):
        """Parse command line arguments and prompt for any missing values.

        :return: A dictionary with the following keys:
            branch, confirm
        """
        args = {}
        args['branch'] = (self.parsed_args.branch
                          if self.parsed_args.branch is not None else
                          self.repo.active_branch.name)
        # Default to config value unless otherwise specified
        args['confirm'] = (self.configs.UNSET_TEMPLATE_CONFIRMATION_PROMPT
                           if self.parsed_args.confirm is None else
                           self.parsed_args.confirm)
        return args

    def run(self):
        args = self.get_args()
        branch = args['branch']
        # Get branch config path, print and exit if non-existent
        branch_config_file = self.configs.get_config(
            f'includeif.onbranch:{branch}.path',
            local=True, includes=True
        )
        if branch_config_file is None:
            self.print(f'Branch {branch} does not have an associated config file.')
            return
        # Confirmation prompt
        if args['confirm']:
            confirmation = cmd.prompt(
                'Unset Template? (y/n)',
                f'Unset commit template for {branch}?',
                default_val='n', validate_function=cmd.validate_yn
            )
            if not confirmation:
                return
        # Verify config file exists
        branch_config_path = os.path.join(self.repo.git_dir, branch_config_file)
        if not os.path.exists(branch_config_path):
            self.print(f'Config file {branch_config_file} not found, unsetting include...')
            self.unset_includeif_onbranch_path(branch)
            self.print_success('Config include removed.', '')
            return
        # Unset commit.template in branch config file
        self.print(f'Unsetting commit.template config for {branch}...')
        commit_template_file = self.configs.get_config(
            'commit.template', file=branch_config_path
        )
        if commit_template_file is None:
            self.print(f'commit.template not configured for branch {branch}.', '')
            return
        self.configs.unset_config(
            'commit.template', file=branch_config_path
        )
        self.print_success('commit.template config unset.', '')
        # Delete commit template
        repo_root_dir = os.path.dirname(self.repo.git_dir)
        commit_template_path = os.path.join(repo_root_dir, commit_template_file)
        if os.path.exists(commit_template_path):
            self.print(f'Deleting commit template file {commit_template_file}...')
            os.remove(commit_template_path)
            self.print_success('Commit template file removed.', '')
        else:
            self.print('Commit template file already removed.')
        # If branch config is now empty, delete the file and unset includeIf
        if os.stat(branch_config_path).st_size == 0:
            self.print(f'Removing empty branch config file and unsetting include...')
            self.unset_includeif_onbranch_path(branch)
            os.remove(branch_config_path)
            self.print_success(f'Empty branch config removed.', '')

    # Helper Methods

    def unset_includeif_onbranch_path(self, branch):
        self.configs.unset_config(
            f'includeif.onbranch:{branch}.path',
            file=self.configs.CONFIG_PATH
        )
